# Copyright 2023 WHERE TRUE Technologies.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FASTA file reader."""
import os

import warnings

from biobear.reader import Reader
from biobear.compression import Compression

from .biobear import _ExonReader


class FastaReader(Reader):
    """FASTA file reader."""

    def __init__(
        self, path: os.PathLike, compression: Compression = Compression.INFERRED
    ):
        """Read a fasta file.

        Args:
            path (Path): Path to the fasta file.

        Kwargs:
            compression (Compression): Compression type of the file. Defaults to
                Compression.INFERRED.

        """
        warnings.warn(
            "The FastaReader class is deprecated and will be removed in a future release. "
            "Please use BioBearSessionContext.read_fasta instead.",
            DeprecationWarning,
        )
        self.compression = compression.infer_or_use(path)

        if self.compression == Compression.GZIP:
            self._fasta_reader = _ExonReader(str(path), "FASTA", "GZIP")
        else:
            self._fasta_reader = _ExonReader(str(path), "FASTA", None)

    @property
    def inner(self):
        """Return the inner reader."""
        return self._fasta_reader



# Copyright 2023 WHERE TRUE Technologies.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BAM File Readers."""

import os

import pyarrow as pa
import warnings

from biobear.reader import Reader
from .biobear import _BamIndexedReader, _ExonReader


class BamReader(Reader):
    """A BAM File Reader."""

    def __init__(self, path: os.PathLike):
        """Initialize the BamReader.

        Args:
            path (Path): Path to the BAM file.

        """
        warnings.warn(
            "The BamReader class is deprecated and will be removed in a future release. "
            "Please use BioBearSessionContext.read_bam_file instead.",
            DeprecationWarning,
        )
        self._bam_reader = _ExonReader(str(path), "BAM", None)

    @property
    def inner(self):
        """Return the inner reader."""
        return self._bam_reader


class BamIndexedReader(Reader):
    """An Indexed BAM File Reader."""

    def __init__(self, path: os.PathLike):
        """Initialize the BamIndexedReader.

        Args:
            path (Path): Path to the BAM file.
            index (Path): Path to the BAM index file.

        """
        warnings.warn(
            "The BamIndexedReader class is deprecated and will be removed in a future release. "
            "Please use BioBearSessionContext.read_bam_file instead.",
            DeprecationWarning,
        )
        self._bam_reader = _BamIndexedReader(str(path))

    @property
    def inner(self):
        """Return the inner reader."""
        return self._bam_reader

    def query(self, region: str) -> pa.RecordBatchReader:
        """Query the BAM file and return an Arrow RecordBatchReader.

        Args:
            region: A region in the format "chr:start-end".

        """
        return self._bam_reader.query(region)



# Use biobear to count the number of records all the files in the data directory

from pathlib import Path
from biobear import FastaReader

# DATA = Path.home() / "data" / "uniref10.fasta.split"
DATA = Path.home() / "data" / "uniref10.fasta.split" / "uniref10.part_001.fasta"

total_records = 0

for batch in FastaReader(DATA).to_arrow():
    total_records += len(batch)

print(total_records)



"""MzML File Reader."""

import os

from biobear.compression import Compression
from biobear.reader import Reader

from .biobear import _ExonReader


class MzMLReader(Reader):
    """A MzML File Reader."""

    def __init__(
        self, path: os.PathLike, compression: Compression = Compression.INFERRED
    ):
        """Initialize the MzMLReader.

        Args:
            path: The path to the MzML file.
        """

        self.compression = compression.infer_or_use(path)

        if self.compression == Compression.GZIP:
            self._reader = _ExonReader(str(path), "MZML", "GZIP")
        else:
            self._reader = _ExonReader(str(path), "MZML", None)

    def to_polars(self):
        """Read the MZML file and return a polars DataFrame."""
        raise RuntimeError(
            "The polars library is not yet supported for MzML files. "
            "Consider using the session to select the fields you need."
        )

    @property
    def inner(self):
        """Return the inner reader."""
        return self._reader



# Copyright 2024 WHERE TRUE Technologies.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyarrow import Table, RecordBatchStreamReader
from typing import Optional
import enum

POLARS_INSTALLED = False
try:
    import polars as pl

    POLARS_INSTALLED = True
except ImportError:
    pass

class FileCompressionType(enum.Enum):
    """The type of compression used for a file."""

    GZIP = 0
    BGZIP = 1
    NONE = 2

class CRAMReadOptions:
    """Options for reading CRAM data."""
    def __init__(
        self,
        /,
        region: Optional[str] = None,
        fasta_reference: Optional[str] = None,
    ) -> None: ...

class FCSReadOptions:
    """Options for reading FCS data."""
    def __init__(
        self,
        /,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class HMMDomTabReadOptions:
    """Options for reading HMM DomTab data."""
    def __init__(
        self,
        /,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class MzMLReadOptions:
    """Options for reading mzML data."""
    def __init__(
        self,
        /,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class GenBankReadOptions:
    """Options for reading GenBank data."""
    def __init__(
        self,
        /,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class GTFReadOptions:
    """Options for reading GTF data."""
    def __init__(
        self,
        /,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class FASTAReadOptions:
    """Options for reading FASTA data."""
    def __init__(
        self,
        /,
        file_extension: Optional[str] = None,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class FASTQReadOptions:
    """Options for reading FASTQ data."""
    def __init__(
        self,
        /,
        file_extension: Optional[str] = None,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class VCFReadOptions:
    """Options for reading VCF data."""
    def __init__(
        self,
        /,
        region: Optional[str] = None,
        file_compression_type: Optional[FileCompressionType] = None,
        parse_info: bool = False,
        parse_formats: bool = False,
        partition_cols: list[str | None] = None,
    ) -> None: ...

class BCFReadOptions:
    """Options for reading BCF data."""
    def __init__(
        self,
        /,
        region: Optional[str] = None,
    ) -> None: ...

class SAMReadOptions:
    """Options for reading SAM data."""
    def __init__(
        self,
    ) -> None: ...

class BAMReadOptions:
    """Options for reading BAM data."""
    def __init__(
        self,
        /,
        region: Optional[str] = None,
    ) -> None: ...

class BEDReadOptions:
    """Options for reading BED data."""
    def __init__(
        self,
        /,
        file_compression_type: Optional[FileCompressionType] = None,
    ) -> None: ...

class BigWigReadOptions:
    """Options for reading BigWig data."""
    def __init__(
        self,
        /,
        zoom: Optional[int] = None,
        region: Optional[str] = None,
    ) -> None: ...

class GFFReadOptions:
    """Options for reading GFF data."""
    def __init__(
        self,
        /,
        file_extension: Optional[str] = None,
        file_compression_type: Optional[FileCompressionType] = None,
        region: Optional[str] = None,
    ) -> None: ...

class ExecutionResult:
    """The result of an execution."""
    def to_arrow(self) -> Table:
        """Converts the result to an Arrow Table."""
    def to_arrow_record_batch_reader(self) -> RecordBatchStreamReader:
        """Converts the result to an Arrow RecordBatchStreamReader."""

    if POLARS_INSTALLED:
        def to_polars(self) -> pl.DataFrame:
            """Converts the result to a Polars DataFrame."""

class BioBearSessionContext:
    def __init__(self) -> None: ...
    def read_fastq_file(
        self, file_path: str, *, options: Optional[FASTQReadOptions] = None
    ) -> ExecutionResult:
        """Reads one or more FASTQ files and returns an ExecutionResult."""
    def read_fasta_file(
        self, file_path: str, *, options: Optional[FASTAReadOptions] = None
    ) -> ExecutionResult:
        """Reads one or more FASTA files and returns an ExecutionResult."""
    def read_vcf_file(
        self, file_path: str, *, options: Optional[VCFReadOptions] = None
    ) -> ExecutionResult:
        """Reads one or more VCF files and returns an ExecutionResult."""
    def read_bcf_file(
        self, file_path: str, *, options: Optional[BCFReadOptions] = None
    ) -> ExecutionResult:
        """Reads one or more BCF files and returns an ExecutionResult."""
    def read_sam_file(
        self, file_path: str, *, options: Optional[SAMReadOptions] = None
    ) -> ExecutionResult:
        """Reads a SAM file and returns an ExecutionResult."""
    def read_bam_file(
        self, file_path: str, *, options: Optional[BAMReadOptions] = None
    ) -> ExecutionResult:
        """Reads a BAM file and returns an ExecutionResult."""
    def read_bed_file(
        self, file_path: str, *, options: Optional[BEDReadOptions] = None
    ) -> ExecutionResult:
        """Reads a BED file and returns an ExecutionResult."""
    def read_bigwig_file(
        self, file_path: str, *, options: Optional[BigWigReadOptions] = None
    ) -> ExecutionResult:
        """Reads a BigWig file and returns an ExecutionResult."""
    def read_gff_file(
        self, file_path: str, *, options: Optional[GFFReadOptions] = None
    ) -> ExecutionResult:
        """Reads a GFF file and returns an ExecutionResult."""
    def read_gtf_file(
        self, file_path: str, *, options: Optional[GTFReadOptions] = None
    ) -> ExecutionResult:
        """Reads a GTF file and returns an ExecutionResult."""
    def read_mzml_file(
        self, file_path: str, *, options: Optional[MzMLReadOptions] = None
    ) -> ExecutionResult:
        """Reads a mzML file and returns an ExecutionResult."""
    def read_genbank_file(
        self, file_path: str, *, options: Optional[GenBankReadOptions] = None
    ) -> ExecutionResult:
        """Reads a GenBank file and returns an ExecutionResult."""
    def read_cram_file(
        self, file_path: str, *, options: Optional[CRAMReadOptions] = None
    ) -> ExecutionResult:
        """Reads a CRAM file and returns an ExecutionResult."""
    def read_fcs_file(
        self, file_path: str, *, options: Optional[FCSReadOptions] = None
    ) -> ExecutionResult:
        """Reads a FCS file and returns an ExecutionResult."""
    def sql(self, query: str) -> ExecutionResult:
        """Executes a SQL query and returns an ExecutionResult."""
    def execute(self, query: str) -> None:
        """Executes a SQL query."""

def connect() -> BioBearSessionContext:
    """Connect to the BioBear server and return a session context.

    Note:
        This function is deprecated. Use `new_session` instead.

    Returns:
        BioBearSessionContext: A session context for interacting with the BioBear server.

    """

def new_session() -> BioBearSessionContext:
    """Create a new session context for interacting with the BioBear server.

    Returns:
        BioBearSessionContext: A session context for interacting with the BioBear server.

    """



"""GFF File Reader."""

import os

import pyarrow as pa
import pyarrow.dataset as ds

from biobear.compression import Compression
from biobear.reader import Reader

from .biobear import _ExonReader


class GFFReader(Reader):
    """A GFF File Reader."""

    def __init__(
        self, path: os.PathLike, compression: Compression = Compression.INFERRED
    ):
        """Initialize the GFFReader.

        Args:
            path: The path to the GFF file.
        """
        self.compression = compression.infer_or_use(path)

        if self.compression == Compression.GZIP:
            self._gff_reader = _ExonReader(str(path), "GFF", "GZIP")
        else:
            self._gff_reader = _ExonReader(str(path), "GFF", None)

    def to_polars(self):
        """Read the GFF file and return a polars DataFrame."""
        try:
            import polars as pl
        except ImportError as import_error:
            raise ImportError(
                "The polars library is required to convert a GFF file "
                "to a polars DataFrame."
            ) from import_error

        return pl.from_arrow(self.to_arrow().read_all())

    def to_arrow(self) -> pa.RecordBatchReader:
        """Convert the GFF reader to an arrow batch reader."""
        return self._gff_reader.to_pyarrow()

    def to_arrow_scanner(self) -> ds.Scanner:
        """Convert the GFF reader to an arrow scanner."""
        return ds.Scanner.from_batches(self.to_arrow())

    @property
    def inner(self):
        """Return the inner reader."""
        return self._gff_reader



"""Abstract Reader class for reading data from a file or stream."""

from abc import ABC, abstractmethod

import pyarrow as pa
import pyarrow.dataset as ds


class Reader(ABC):
    """An abstract base class (ABC) representing a reader.

    The class defines basic functionalities for conversion, but the specifics must be
    implemented in a subclass.
    """

    @property
    @abstractmethod
    def inner(self):
        """Abstract property for the inner reader.

        Returns:
            The inner reader. The type of the reader is defined by the specific
            subclass.
        """

    def to_pandas(self):
        """Convert the inner data to a Pandas DataFrame.

        Returns:
            pd.DataFrame: The converted data in a Pandas DataFrame.

        Raises:
            ImportError: If the 'pandas' package is not installed.
        """
        try:
            import pandas as pd  # noqa

            return self.to_arrow_scanner().to_table().to_pandas()
        except ImportError as import_error:
            raise ImportError(
                "The 'pandas' package is required to use the to_pandas method."
            ) from import_error

    def to_polars(self):
        """Convert the inner data to a Polars DataFrame.

        This method first converts the inner reader's data to an Arrow table,
        then to a Python dictionary, and finally to a Polars DataFrame.

        Returns:
            pl.DataFrame: The converted data in a Polars DataFrame.

        Raises:
            ImportError: If the 'polars' package is not installed.
        """
        try:
            import polars as pl
        except ImportError as import_error:
            raise ImportError(
                "The 'polars' package is required to use the to_polars method."
            ) from import_error

        pydict = self.to_arrow_scanner().to_table().to_pydict()
        return pl.from_dict(pydict)

    def to_arrow_scanner(self) -> ds.Scanner:
        """Convert the inner data to an Arrow scanner.

        This method first converts the inner reader's data to Arrow batches,
        and then forms a scanner from these batches.

        Returns:
            ds.Scanner: The converted data in an Arrow scanner.
        """
        return ds.Scanner.from_batches(self.to_arrow())

    def to_arrow(self) -> pa.RecordBatchReader:
        """Convert the inner data to an Arrow record batch reader.

        If the inner reader is exhausted, this method raises an exception.
        Otherwise, it converts the inner reader's data to an Arrow record batch.

        Returns:
            pa.RecordBatchReader: The converted data in an Arrow record batch reader.

        Raises:
            StopIteration: If the inner reader is exhausted.
        """
        if self.inner.is_exhausted():
            raise StopIteration("The reader is exhausted.")

        return self.inner.to_pyarrow()
